use std::ops::Range;

use color_space::{Hsv, Rgb};
use either::Either;

use super::{
    config::{DisplayConfig, LedStripConfig},
    led_color::LedColor,
};

#[derive(Clone)]
pub struct Screenshot {
    bitmap: Vec<u8>,
    config: DisplayConfig,
}

impl Screenshot {
    pub fn new(bitmap: Vec<u8>, config: DisplayConfig) -> Self {
        Self { bitmap, config }
    }

    pub async fn get_top_colors(&self) -> anyhow::Result<Vec<LedColor>> {
        self.get_x_colors(XPosition::Top, self.config.top_led_strip)
            .await
    }
    pub async fn get_bottom_colors(&self) -> anyhow::Result<Vec<LedColor>> {
        self.get_x_colors(XPosition::Bottom, self.config.bottom_led_strip)
            .await
    }

    pub fn get_top_of_led_start_at(&self) -> usize {
        self.config.top_led_strip.global_start_position
    }
    pub fn get_top_of_led_end_at(&self) -> usize {
        self.config.top_led_strip.global_end_position
    }

    async fn get_x_colors(
        &self,
        position: XPosition,
        strip_config: LedStripConfig,
    ) -> anyhow::Result<Vec<LedColor>> {
        let bitmap = &self.bitmap;
        let number_of_leds = strip_config
            .global_start_position
            .abs_diff(strip_config.global_end_position)
            + 1;
        if number_of_leds == 0 {
            return Ok(vec![]);
        }
        let cell_size_x = self.config.display_width / number_of_leds;
        let cell_size_y = self.config.display_height / 8;
        let cell_size = cell_size_x * cell_size_y;
        let y_range = match position {
            XPosition::Top => 20..cell_size_y + 20,
            XPosition::Bottom => {
                self.config.display_height - 20 - cell_size_y..self.config.display_height - 20
            }
        }
        .step_by(5);

        let x_range = if strip_config.global_start_position < strip_config.global_end_position {
            Either::Left(strip_config.global_start_position..=strip_config.global_end_position)
        } else {
            Either::Right(
                (strip_config.global_end_position..=strip_config.global_start_position).rev(),
            )
        };

        let mut colors = Vec::new();
        let stride = bitmap.len() / self.config.display_height;

        for pos in x_range {
            let mut r = 0.0;
            let mut g = 0.0;
            let mut b = 0.0;
            let mut count = 0;
            for x in (pos * cell_size_x..(pos + 1) * cell_size_x).step_by(5) {
                for y in y_range.to_owned() {
                    let i = stride * y + 4 * x;
                    r += bitmap[i + 2] as f64;
                    g += bitmap[i + 1] as f64;
                    b += bitmap[i] as f64;
                    count+=1;
                }
            }
            let rgb = Rgb::new(
                r / count as f64,
                g / count as f64,
                b / count as f64,
            );
            let hsv = Hsv::from(rgb);
            // info!("HSV: {:?}", [hsv.h, hsv.s, hsv.v]);
            let color = LedColor::from_hsv(hsv.h, hsv.s, hsv.v);
            // paris::info!("color: {:?}", color.get_rgb());
            colors.push(color);
        }
        return Ok(colors);
    }

    pub async fn to_webp_base64(&self) -> String {
        let bitmap = &self.bitmap;
        let mut bitflipped =
            Vec::with_capacity(self.config.display_width * self.config.display_height * 3);
        let stride = bitmap.len() / self.config.display_height;

        for y in 0..self.config.display_height {
            for x in 0..self.config.display_width {
                let i = stride * y + 4 * x;
                bitflipped.extend_from_slice(&[bitmap[i + 2], bitmap[i + 1], bitmap[i]]);
            }
        }

        let webp_memory = webp::Encoder::from_rgb(
            bitflipped.as_slice(),
            self.config.display_width as u32,
            self.config.display_height as u32,
        )
        .encode(100.0);
        return base64::encode(&*webp_memory);
    }
}

enum XPosition {
    Top,
    Bottom,
}
